local i4, i2, i1, u4, u2, u1 = mem.i4, mem.i2, mem.i1, mem.u4, mem.u2, mem.u1
local mmver = offsets.MMVersion

evt = {}
Evt = evt

local _KNOWNGLOBALS = vars, Vars, mapvars, MapVars, Game, Party, Map

local function MakeEventsTable()
	local ret = events.new()
	local m = getmetatable(ret)
	local get = m.__index
	local set = m.__newindex
	local defcall = ret.call
	function m.__index(t, a)
		if type(a) ~= "number" then
			return rawget(t, a)
		end
		local function index(_, a1)
			if type(a1) == "number" then
				return get(t, a*256 + a1)
			else
				return get(t, a*256)[a1]
			end
		end
		local function newindex(_, a1, v)
			set(t, a*256 + a1, v)
		end
		local function call(_, ...)
			return defcall(a*256, ...)
		end
		return setmetatable({}, {__index = index, __newindex = newindex, __call = call})
	end
	function m.__newindex(t, a, v)
		if type(a) ~= "number" then
			return rawset(t, a, v)
		end
		set(t, a*256, v)
	end
	return ret
end

local ResetMapStr
local ClearMapStr
local UpdateEventJustHint

local CurMapScripts = {}
local GlobalScripts = {}
local MayShow = 0  -- 0 in loading screen, 1 after the game was loaded
local WasInGame

local function OnLeaveMap()
	events.cocall("LeaveMap")
	events.RemoveFiles(CurMapScripts)
	evt.global.RemoveFiles(CurMapScripts)
	internal.TimersLeaveMap(CurMapScripts)
	CurMapScripts = {}
	internal.MonstersRestore()
	ClearMapStr()
	internal.MapName = nil
end
internal.OnLeaveMap = OnLeaveMap

local function OnLeaveGame()
	MayShow = 0
	if internal.InGame then
		events.cocall("LeaveGame")
		internal.TimersLeaveGame()
		internal.InGame = false
	end
	table.copy(CurMapScripts, GlobalScripts)
	events.RemoveFiles(GlobalScripts)
	GlobalScripts = {}
	CurMapScripts = {}
	internal.SaveGameData = nil
	ClearMapStr()
	internal.MapName = nil
end
internal.OnBeforeLoadGame = OnLeaveGame
internal.OnExitToMainMenu = OnLeaveGame

function internal.BeforeMapLoad()
	MayShow = 0
	local sgd = internal.SaveGameData or {}
	internal.SaveGameData = sgd
	sgd.Vars = sgd.Vars or {}
	sgd.Maps = sgd.Maps or {}
	vars = sgd.Vars
	Vars = vars
	if mmver == 6 then
		evt.MazeInfo = nil
	end
	local MapName = mem.string(offsets.MapName, 20):lower()
	mapvars = sgd.Maps[MapName] or {}
	MapVars = mapvars
	sgd.Maps[MapName] = mapvars
	WasInGame = internal.InGame
	if not WasInGame then
		internal.InGame = true
		evt.Global = MakeEventsTable()
		evt.global = evt.Global
		if mmver > 6 then
			evt.CanShowTopic = events.new()
		end
		for f in path.find("Scripts/Global/*.lua") do
			local chunk, err = loadfile(f)
			if chunk == nil then
				debug.ErrorMessage(err)
			else
				GlobalScripts[debug.FunctionFile(chunk)] = true
				coroutine.resume2(coroutine.create(chunk))
			end
		end
	end
	events.cocall("BeforeLoadMap")
end

function internal.OnLoadMap()
	local MapName = mem.string(offsets.MapName, 20):lower()
	internal.MapName = MapName  -- this is needed because my OnMapLeave works in MM6 when map name is already changed
	internal.UpdateMapRefilled()
	evt.Map = MakeEventsTable()
	evt.map = evt.Map
	evt.Hint = {}
	evt.hint = evt.Hint
	if mmver == 6 then
		evt.MazeInfo = nil
	end
	ResetMapStr()
	internal.StartTimers()
	internal.LoadMonsterIds()
	CurMapScripts = {}
	for f in path.find("Scripts/Maps/*."..path.SetExt(MapName, ".lua")) do
		local chunk, err = loadfile(f)
		if chunk == nil then
			debug.ErrorMessage(err)
		else
			CurMapScripts[debug.FunctionFile(chunk)] = true
			coroutine.resume2(coroutine.create(chunk))
		end
	end
	for f in path.find("Scripts/Maps/"..path.SetExt(MapName, ".lua")) do
		local chunk, err = loadfile(f)
		if chunk == nil then
			debug.ErrorMessage(err)
		else
			CurMapScripts[debug.FunctionFile(chunk)] = true
			coroutine.resume2(coroutine.create(chunk))
		end
	end
	events.cocall("LoadMap", WasInGame)
	UpdateEventJustHint()
	MayShow = 1
end

function internal.AfterLoadMap()
	events.cocall("AfterLoadMap")
end

----------- Hint, MazeInfo

local TextBuffer = internal.TextBuffer

local function GetHouseHint(id)
	if id then
		return Game.Houses[id].Name
	end
end

function internal.OnGetEventHint(evtId)
	local h = evt.house and (events.call("GetEventHint", evtId) or evt.hint[evtId] or GetHouseHint(evt.house[evtId]))
	if h then
		mem.copy(TextBuffer, h, #h + 1)
	end
	return h and 1 or 0
end

if mmver == 6 then
	function internal.OnGetMazeInfo()
		local h = events.call("GetMazeInfo") or evt.MazeInfo
		if h then
			mem.copy(TextBuffer, h, #h + 1)
		end
		return h and 1 or 0
	end
end

----------- UpdateEventJustHint

function UpdateEventJustHint()
	if mmver == 6 or not evt.house then
		-- do nothing
	elseif Map.IsIndoors() then
		for _, a in Map.FacetData do
			local ev = a.Event
			if ev ~= 0 and (evt.house[ev] or evt.map.exists(ev*256)) then
				Map.Facets[a.FacetIndex].IsEventJustHint = false
			elseif ev ~= 0 and evt.hint[ev] then
				Map.Facets[a.FacetIndex].IsEventJustHint = true
			end
		end
	else
		for _, a in Map.Models do
			for _, a in a.Facets do
				local ev = a.Event
				if ev ~= 0 and (evt.house[ev] or evt.map.exists(ev*256)) then
					a.IsEventJustHint = false
				elseif ev ~= 0 and evt.hint[ev] then
					a.IsEventJustHint = true
				end
			end
		end
	end
end

----------- CanShowTopic

function internal.CanShowTopic(topic, def)
	local ret = evt.CanShowTopic[topic]()
	if ret == nil then
		return def
	end
	return ret and 1 or 0
end

----------- Str

local strbuf = {}

local function str_index(t, a)
	if a >= 500 or a < 0 then
		error("evt.str index out of bounds [0..499]", 2)
	end
	return mem.string(offsets.MapStrBuf + u4[offsets.MapStrOffsets + a*4])
end

local function str_newindex(t, a, v)
	if a >= 500 or a < 0 then
		error("evt.str index out of bounds [0..499]", 2)
	end
	local buf = strbuf[a]
	if buf then
		mem.free(buf)
	end
	buf = mem.malloc(#v + 1)
	strbuf[a] = buf
	mem.copy(buf, v, #v + 1)
	u4[offsets.MapStrOffsets + a*4] = buf - offsets.MapStrBuf
end

function ClearMapStr()
	if strbuf then
		for i, v in pairs(strbuf) do
			mem.free(v)
		end
	end
	strbuf = nil
	evt.str = nil
	evt.Str = nil
	evt.house = nil
	evt.House = nil
end

function ResetMapStr()
	strbuf = {}
	evt.str = setmetatable({}, {__index = str_index, __newindex = str_newindex})
	evt.Str = evt.str
	evt.house = {}
	evt.House = evt.house
end

----------- AfterProcessEvent

local GlobalEventInfo, TargetObj
local AsyncProc, AsyncTrue, AsyncPlayer, AsyncCurrentPlayer, AsyncTargetObj

function internal.AfterProcessEvent(evtId, seq, globalEventInfo, targetObj, param, retseq)
	if evt.house == nil then  -- OnLoadMap wasn't executed
		return
	end
	local old1, old2, old3, old4 = GlobalEventInfo, TargetObj, evt.Player, evt.CurrentPlayer
	GlobalEventInfo, TargetObj, MayShow, evt.Player = globalEventInfo, targetObj, param, u4[offsets.CurrentPlayer] - 1
	if evt.Player < 0 then
		evt.Player = math.random(offsets.PlayersCount) - 1
	end
	evt.CurrentPlayer = evt.Player
	local f = AsyncProc
	AsyncProc = nil
	if evtId ~= 0x7FFF then
		local ev = (globalEventInfo == 0) and evt.map or evt.global
		local evn = evtId*256 + seq
		if ev.exists(evn) then
			ev.cocall(evn)
		elseif globalEventInfo == 0 and evt.house[evtId] then
			evt.EnterHouse(evt.house[evtId])
		end
		events.cocall((globalEventInfo == 0) and "EvtMap" or "EvtGlobal", evtId, seq)
	elseif f then
		evt.Player, evt.CurrentPlayer = AsyncPlayer, AsyncCurrentPlayer
		if type(f) == "thread" then
			coroutine.resume2(f, retseq == AsyncTrue)
		else
			pcall2(f, retseq == AsyncTrue)
		end
	end
	GlobalEventInfo, TargetObj, evt.Player, evt.CurrentPlayer = old1, old2, old3, old4
end

--------- Some Functions

function evt.InGlobal()
	return GlobalEventInfo and GlobalEventInfo ~= 0 and true or false
end

function evt.ForPlayer(n)
	if type(n) == "string" then
		evt.Player = evt.Players[n]
	else
		evt.Player = n
	end
	return evt
end

--------- Call Events

local CmdDef = {}
local CmdStructs = {}
local CmdNames = {}
local CmdJump = {}
local CmdFieldsOrder = {}
local BitsArray = {}
local DeclareCmd = {}

local EvtBuf = internal.EvtBuf
local BufPtr, LineN

local function FillLine()
	local p = offsets.CurrentEvtLines + LineN*12
	u4[p] = 0x7FFF
	u4[p + 4] = LineN
	u4[p + 8] = BufPtr - EvtBuf
	LineN = LineN + 1
end

local Jump, JumpY, TextName -- arguments for MakeCmd

local function MakeCmd(name, num, f, invis)
	Jump, TextName = nil, nil
	local offs0
	local cmdStruct = mem.struct(function(define, ...)
		offs0 = define.offsets
		return f(define, ...)
	end)
	CmdStructs[num] = cmdStruct
	CmdNames[num] = name
	local textName = TextName
	local def = CmdDef[num]
	local order = {}
	CmdFieldsOrder[num] = order
	CmdJump[num] = Jump
	local jump, jumpTrue = Jump, (JumpY and 3 or 2)

	if cmdStruct["?size"] ~= #def - ((num == 6 or num == 0xD) and 1 or 0) then
		error(("Cmd%d size(%d) doesn't match CmdDef size(%d)"):format(num, cmdStruct["?size"], #def - (num == 6 and 1 or 0)), 2)
	end

	-- fill order
	do
		local offs = {}
		for a, b in pairs(offs0) do
			offs[b] = a
		end
		local j = 0
		local s
		for i = 0, cmdStruct["?size"] - 1 do
			s = offs[i]
			if s and s ~= jump then
				j = j + 1
				order[j] = s
			end
		end
	end

	DeclareCmd[num] = function(t)
		u2[BufPtr + 1] = 0x7FFF
		u1[BufPtr + 3] = LineN
		u1[BufPtr + 4] = num
		FillLine()
		local len = #def
		mem.copy(BufPtr + 5, def, len)
		if t then
			local a = cmdStruct:new(BufPtr + 5)
			local kn
			for k,v in pairs(t) do
				kn = order[k] or k
				a[kn] = v
				if kn == textName then
					len = cmdStruct["?size"] + #v
				end
			end
		end
		u1[BufPtr] = len + 4
		BufPtr = BufPtr + len + 5
	end

	if not invis then
		evt[name] = function(t, ...)
			if t == nil or type(t) ~= "table" then
				t = {t, ...}
			end
			if jump then
				t[jump] = 3
			end

			BufPtr = EvtBuf
			LineN = 0
			DeclareCmd[0]()  -- for buggy commands like 0x1A
			DeclareCmd[num](t)
			DeclareCmd[1]()
			DeclareCmd[1]()  -- for jump

			if num == 3 then
				Game.LoadSound(t.Id or 0)
			end

			u4[offsets.CurrentEvtLinesCount] = LineN
			if GlobalEventInfo or u4[offsets.GlobalEventInfo] == 0 then
				u4[offsets.GlobalEventInfo] = GlobalEventInfo or (Game.CurrentScreen == 4 or Game.CurrentScreen == 13) and 1 or 0
			end
			u4[offsets.EvtTargetObj] = TargetObj or 0
			local player = evt.Player
			if not player or player == evt.Players.Current then
				player = evt.CurrentPlayer or evt.Players.Current
			end
			local ret = mem.call(internal.CallProcessEvent, 0, 0x7FFF, 0, player, MayShow)

			if ret == 0 then -- async command
				AsyncTrue = jumpTrue
				AsyncProc = t.OnDone
				AsyncPlayer, AsyncCurrentPlayer = evt.Player, evt.CurrentPlayer
				local c = not AsyncProc and not t.NoYield and coroutine.running()
				if c then
					AsyncProc = c
					return coroutine.yield()
				end
			end
			return (ret == jumpTrue)
		end
	end
end

--------- Decompile

local VarNumToStr, PlayerToStr

function evt.Decompile(fileName, funcMode, outFile)

	local funcMode = funcMode or 0
	local insert = table.insert
	local evtStr
	local IsGlobal = path.name(fileName):lower() == "global.evt"

	local function DecompileStr(buf, size)
		local p = buf
		evtStr = {}
		size = p + size
		local i = 1
		local s
		while p < size do
			s = mem.string(p)
			evtStr[i] = s
			p = p + #s + 1
			i = i + 1
		end
		if evtStr[2] == nil then  -- #0 are changed into #13#10 by extractors
			p = buf
			i = 1
			for j = buf, size - 1 do
				if u1[j] == 13 then
					u1[j] = 0
					evtStr[i] = mem.string(p)
					p = j + 2
					i = i + 1
				end
			end
		end
		for i, v in ipairs(evtStr) do
			if v:sub(1, 1) ~= '"' then
				evtStr[i] = ('"%s"'):format(v)
			end
		end
	end

	local function GetFromArray(arr, n, name)
		if n >= arr.low and n <= arr.high then
			return '"'..(name and arr[n][name] or arr[n])..'"'
		end
	end

	local function GetStr(n)
		return IsGlobal and GetFromArray(Game.NPCText, n) or evtStr[n + 1]
	end

	local function FindConst(name, v)
		v = table.find(const[name], v)
		if v then
			return "const."..name.."."..v
		end
	end
	
	local function GetVarNumComment(k, v)
		if k == '"ClassIs"' then
			return GetFromArray(Game.ClassNames, v)
		elseif k == '"Awards"' then
			return GetFromArray(Game.AwardsTxt, v)
		elseif k == '"QBits"' then
			local ret = GetFromArray(Game.QuestsTxt, v)
			return ret ~= '""' and ret or nil
		elseif k == '"Inventory"' then
			return GetFromArray(Game.ItemsTxt, v, 'Name')
		elseif k == '"MainCondition"' then
			return FindConst("Condition", v)
		elseif k == '"AutonotesBits"' then
			return GetFromArray(Game.AutonoteTxt, v)
		elseif k == '"NPCs"' then
			return GetFromArray(Game.NPC, v, 'Name')
		elseif k == '"HasNPCProfession"' then
		end
	end
	
	local function DecompileCmd(p, wasShowTopic)
		local label = u1[p + 3]
		local num = u1[p + 4]
		local struct = CmdStructs[num]
		if not struct then
			return "unknown command: "..num, false
		end
		struct = struct:new(p + 5)
		local str = {}
		local showTopic = (CmdNames[num]:sub(1, #"CanShowTopic.") == "CanShowTopic.")
		if wasShowTopic ~= nil and showTopic ~= wasShowTopic then
			str[1] = "\r\n"
		end
		local order = CmdFieldsOrder[num]
		local func = funcMode == 2 or (funcMode == 1 and #order == 1)
		if num == 0x23 then
			insert(str, (label < 10 and "  %s:  Player = %s" or "  %s: Player = %s"):format(label, PlayerToStr[struct.Player] or struct.Player))
		elseif (num == 4 or num == 5) and label == 0 then
			local v = struct.Str
			local comment = evtStr[v + 1]
			if comment then
				insert(str, ("      %s = str[%s]  -- %s"):format(CmdNames[num], v, comment))
			else
				insert(str, ("      %s = str[%s]"):format(CmdNames[num], v))
			end
		else
			insert(str, (label < 10 and "  %s:  %s  " or "  %s: %s  "):format(label, CmdNames[num]))
			insert(str, func and "(" or "{")
			local comment
			local jump = CmdJump[num]
			local hasParams
			
			-- comment

			if num == 0x02 then
				local i = struct.Id
				if i == 600 then
					comment = (mmver == 7) and "Win Good" or "Win"
				elseif i == 601 then
					comment = (mmver == 7) and "Win Evil" or "Lose"
				else
					comment = GetFromArray(Game.Houses, struct.Id, 'Name')
				end
			elseif num == 0x0F and struct.State == 2 then
				comment = "switch state"
			elseif num == 0x15 then
			elseif num == 0x16 then
				comment = GetFromArray(Game.NPC, struct.NPC, 'Name')
			elseif num == 0x1A then
				local old = IsGlobal
				if mmver == 8 then
					IsGlobal = true
				end
				comment = GetStr(struct.Question)
				local a1, a2 = GetStr(struct.Answer1), GetStr(struct.Answer2)
				if a1 or a2 then
					comment = comment and comment.." (" or "("
					if a1 == nil then
						a1, a2 = a2, a1
					end
					comment = comment..a1
					if a2 then
						comment = comment..", "..a2
					end
					comment = comment..")"
				end
				IsGlobal = old
			elseif (num == 0x1F or num == 0x26) and not struct.EachYear and not struct.EachMonth 
						 and not struct.EachWeek and struct.IntervalHalfMins == 0 then
				comment = "each day"
			elseif num == 0x22 then
			elseif num == 0x27 then
			elseif num == 0x28 then
			elseif num == 0x29 then  -- GiveItem
				if struct.Id ~= 0 then
					comment = GetFromArray(Game.ItemsTxt, struct.Id, 'Name')
				end
			elseif num == 0x2B then
			elseif num == 0x2F then
			elseif num == 0x31 then
			elseif num == 0x32 then
			elseif num == 0x3C then
			elseif num == 0x40 then
			elseif num == 0x41 then
			elseif num == 0x44 then
			end

			-- command call 
			for i = 1, #order + (jump and 1 or 0) do
				local a = order[i] or jump
				local v = struct[a]
				if v or a == "On" or a == "Has" or a == "RandomAngle" or a == "Visible" then
					if hasParams then
						insert(str, ", ")
					else
						hasParams = true
					end
					local vn = (a == "VarNum" and (VarNumToStr[v] or v))
									or (a == "Player" and PlayerToStr[v])
									or (num == 0x29 and i == 2 and FindConst("ItemType", v))
									or (a == "DamageType" and FindConst("Damage", v))
									or (a == "Bit" and FindConst(BitsArray[num], v))
					if a == "VarNum" then
						comment = GetVarNumComment(v, struct.Value)
					elseif a == "Str" then
						comment = GetStr(v)
					elseif a == "Event" or a == "NewEvent" then
						comment = GetFromArray(Game.NPCTopic, v)
					elseif a == "HouseId" then
						comment = GetFromArray(Game.Houses, v, 'Name')
					end
					if not func and (not vn or i ~= 1) then
						insert(str, a)
						insert(str, " = ")
					end
					if vn then
						insert(str, vn)
					elseif type(v) == "string" then
						insert(str, '"')
						insert(str, v)
						insert(str, '"')
					elseif a == "Bit" then
						insert(str, ("0x%X"):format(v))
					else
						insert(str, tostring(v))
					end
				end
			end
			insert(str, func and ")" or "}")
			if comment then
				insert(str, '  -- ')
				insert(str, comment)
			end
		end
		return table.concat(str), num == 1 or num == 0x24, showTopic  -- exit or goto
	end

	local function DecompileBuffer(p, size)
		local str = {}

		if evtStr then
			for i, v in ipairs(evtStr) do
				insert(str, ('str[%s] = %s'):format(i - 1, v))
			end
			insert(str, "\n")
		else
			evtStr = {}
		end

		local curEvt
		local s, needLine, showTopic
		size = p + size
		while p < size and p + u1[p] < size do
			local evt = u2[p + 1]
			if evt ~= curEvt then
				showTopic = nil
				if curEvt then
					insert(str, "end\r\n")
				end
				s = "event "..evt
				if IsGlobal and evt >= Game.NPCTopic.low and evt <= Game.NPCTopic.high then
					local s1 = Game.NPCTopic[evt]
					if s1 and s1 ~= "" then
						s = s..(('  -- "%s"'):format(s1))
					end
				end
				insert(str, s)
				curEvt = evt
			elseif needLine then
				insert(str, "")
			end
			s, needLine, showTopic = DecompileCmd(p, showTopic)
			insert(str, s)
			p = p + u1[p] + 1
		end
		if curEvt then
			insert(str, "end\r\n")
		end
		return table.concat(str, "\r\n")
	end

	local s, size, buf

	for f in path.find(path.SetExt(fileName, ".str")) do
		s = io.LoadString(f)
		size = #s
		if size then
			buf = mem.malloc(size + 1)
			mem.copy(buf, s, size + 1)
			DecompileStr(buf, size)
			mem.free(buf)
		end
	end
	s = io.LoadString(fileName)
	if #s < 5 then
		return nil
	end
	size = #s
	buf = mem.malloc(size + 30)
	mem.copy(buf, s, size + 1)  -- +30 and size + 1 are here to be sure no out-of-bounds problems occur
	s = DecompileBuffer(buf, size)
	mem.free(buf)
	if outFile then
		io.SaveString(outFile, s)
	end
	return s
end

--------------------------------------------------------------------------------

do
	CmdDef[0x00] = ""
	CmdDef[0x01] = "\0"
	CmdDef[0x02] = "\0\0\0\0"
	CmdDef[0x03] = "\0\0\0\0\0\0\0\0\0\0\0\0"
	CmdDef[0x04] = "\0"
	CmdDef[0x05] = "\0" -- MM6!
	if mmver == 8 then
		CmdDef[0x06] = "\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\48\0"
	else
		CmdDef[0x06] = "\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\48\0"
	end
	CmdDef[0x07] = "\0"
	CmdDef[0x08] = "\4\0" -- MM6!
	if mmver == 8 then
		CmdDef[0x09] = "\7\0\0\0\0\0"
	else
		CmdDef[0x09] = "\4\0\0\0\0\0"
	end
	CmdDef[0x0A] = "\0\0" -- MM6!
	CmdDef[0x0B] = "\0\0\0\0\0"
	if mmver ~= 6 then
		CmdDef[0x0C] = "\1\0\0" -- (ShowMovie)
	else
		CmdDef[0x0C] = "\0\0\0\0\0\0\0\0\0" -- (?)
	end
	CmdDef[0x0D] = "\0\0\0\0\1\48\0"
	if mmver ~= 6 then
		CmdDef[0x0E] = "\0\0\0\0\0\0\0"
	else
		CmdDef[0x0E] = "\0\0\0\0\0\0"
	end
	CmdDef[0x0F] = "\0\0"
	if mmver ~= 6 then
		CmdDef[0x10] = "\0\0\0\0\0\0"
	else
		CmdDef[0x10] = "\0\0\0\0\0"
	end
	if mmver ~= 6 then
		CmdDef[0x11] = "\0\0\0\0\0\0"
	else
		CmdDef[0x11] = "\0\0\0\0\0"
	end
	if mmver ~= 6 then
		CmdDef[0x12] = "\0\0\0\0\0\0"
	else
		CmdDef[0x12] = "\0\0\0\0\0"
	end
	if mmver ~= 6 then
		CmdDef[0x13] = "\1\1\1\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0"
	else
		CmdDef[0x13] = "\1\1\1\0\0\0\0\0\0\0\0\0\0\0\0"
	end
	CmdDef[0x14] = "\0\0\0\0\1"
	CmdDef[0x15] = "\0\3\5\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0"
	CmdDef[0x16] = "\0\0\0\0"
	CmdDef[0x17] = "\0\0\0\0\0\0\0\0\0"
	if mmver ~= 6 then
		CmdDef[0x18] = "\0\0\0\0\0\0\0\0\0"
	else
		CmdDef[0x18] = "\0\0\0\0\255\255\255\255\0\0\0\0\0"
	end
	CmdDef[0x19] = "\1\0\0\0\0\0"
	CmdDef[0x1A] = "\0\0\0\0\0\0\0\0\0\0\0\0\0"
	CmdDef[0x1B] = "\0\0"
	CmdDef[0x1C] = "\0"
	CmdDef[0x1D] = "\0\0\0\0"
	CmdDef[0x1E] = "\0\0\0\0"
	CmdDef[0x1F] = "\0\0\0\0\0\0\0\0\0\0"
	CmdDef[0x20] = "\1\0\0\0\0"
	CmdDef[0x21] = "\0"
	CmdDef[0x22] = "\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\232\3\0\0\1\0"
	CmdDef[0x23] = "\0"
	CmdDef[0x24] = "\0"
	CmdDef[0x25] = "\0"
	CmdDef[0x26] = "\0\0\0\0\0\0\0\0\0\0"
	CmdDef[0x27] = "\0\0\0\0\0\0\0\0\0"
	CmdDef[0x28] = "\0\0\0\0\0\0\0\0"
	CmdDef[0x29] = "\1\0\0\0\0\0"
	CmdDef[0x2A] = "\0\0\0\0"
	CmdDef[0x2B] = "\0\3\40\0\0\0\0"
	CmdDef[0x2C] = "\0\0\0\0\0\0\0"
	CmdDef[0x2D] = "\0"
	CmdDef[0x2E] = "\0"
	CmdDef[0x2F] = "\0\0\0\0\0\0\0\0"
	CmdDef[0x30] = "\0\0\0\0\0\0\0\0"
	CmdDef[0x31] = "\0\0\0\0\0\0\0\0\0"
	CmdDef[0x32] = "\0\0\0\0\0\0\0\0"
	if mmver == 8 then
		CmdDef[0x33] = "\0\0\0\0\0\0\1\0"
	else
		CmdDef[0x33] = "\0\0\0\0\0\0\0"
	end
	CmdDef[0x34] = CmdDef[0x33]
	CmdDef[0x35] = "\0"
	CmdDef[0x36] = "\0\0\0\0\0\0\0\0"
	CmdDef[0x37] = "\0\0\0\0\0\0\0\0"
	CmdDef[0x38] = "\0\0"
	CmdDef[0x39] = "\0\0\0\0\0\0\0\0\0"
	CmdDef[0x3A] = "\0\0\0\0\1\0\0\0\0"
	if mmver == 8 then
		CmdDef[0x3B] = "\7\18"
	else
		CmdDef[0x3B] = "\4\18"
	end
	CmdDef[0x3C] = "\0\0\0\0\0\0\0\0\1"
	CmdDef[0x3D] = "\0\0\0\0\0\0\0\0"
	CmdDef[0x3E] = "\0\0\0"
	CmdDef[0x3F] = "\0\0\0\0"
	CmdDef[0x40] = "\0\0\0\0\0\0\0"
	CmdDef[0x41] = "\0\0\0\0\0\0"
	CmdDef[0x42] = "\0\0\0\0\220\5\0\0"
	CmdDef[0x43] = "\0\0\0\0\0\0\0\0\0"
	CmdDef[0x44] = "\0\0\0\0\0"
end

--------------------------------------------------------------------------------

(function()

	local types_u1
	do
		mem.struct(function(define)  types_u1 = define.u1  end)
	end
	local function EvtVar(name)
		local a = mmver == 6 and u1 or u2
		return mem.structs.CustomType(name, mmver == 6 and 1 or 2,
			function(o, obj, name, val)
				local p = obj["?ptr"] + o
				if val then
					a[p] = evt.VarNum[val] or val
				else
					val = a[p]
					return VarNumToStr[val] or val --a[p]
				end
			end
		)
	end

	local function TEvtString(name)
		TextName = name
		return mem.structs.CustomType(name, 1,
			function(o, obj, name, val)
				local p = obj["?ptr"] + o
				if val then
					mem.copy(p, val, #val + 1)
				else
					return mem.string(p)
				end
			end
		)
	end

	local function YJump(name)
		Jump = name
		JumpY = true
		return types_u1(name)
	end

	local function NJump(name)
		Jump = name
		JumpY = false
		return types_u1(name)
	end

	---------------------------
	MakeCmd("Cmd00", 0x00, function()
	end, true)

	---------------------------
	MakeCmd("Exit", 0x01, function(define)
		define
		.size = 1
	end, true)

	---------------------------
	MakeCmd("EnterHouse", 0x02, function(define)
		define
		.u4  'Id'  --  of 2DEvent
		 -- 600(dec) = you won
		 -- 601(dec) = you won 2 / you lost
	end)

	---------------------------
	MakeCmd("PlaySound", 0x03, function(define)
		define
		.i4  'Id'
		.i4  'X'
		.i4  'Y'
	end)

	---------------------------
	MakeCmd("Hint", 0x04, function(define)
		define
		.u1  'Str'
	end, true)

	---------------------------
	MakeCmd("MazeInfo", 0x05, function(define)
		define
		.u1  'Str'
	end, true)

	---------------------------
	MakeCmd("MoveToMap", 0x06, function(define)
	-- Notes:
	-- If cancel is pressed, the execution is stopped
	-- If X,Y,Z,Direction,LookAngle,SpeedZ are all 0, the party isn't moved
	-- If HouseId and Icon are 0, the enter dungeon dialog isn't shown

		define
		.i4  'X'
		.i4  'Y'
		.i4  'Z'
		.i4  'Direction'  -- -1 = special case
		.i4  'LookAngle'
		.i4  'SpeedZ'
		if mmver == 8 then
			define.u2  'HouseId'  -- 2DEvent ID
		else
			define.u1  'HouseId'  -- 2DEvent ID
		end
		define.u1  'Icon'
		TEvtString 'Name' -- if starts with "0" => current map
	end)

	---------------------------
	MakeCmd("OpenChest", 0x07, function(define)
		define
		.u1  'Id'
	end)

	---------------------------
	MakeCmd("FaceExpression", 0x08, function(define)
		define
		.u1  'Player'
		.u1  'Frame'
	end)

	---------------------------
	MakeCmd("DamagePlayer", 0x09, function(define)
		define
		.u1  'Player'
		.u1  'DamageType'
		.i4  'Damage'
	end)

	---------------------------
	MakeCmd("SetSnow", 0x0A, function(define)
		define
		.u1  'EffectId'  -- only 00 available
		.b1  'On'
	end)

	---------------------------
	MakeCmd("SetTexture", 0x0B, function(define)
		define
		.i4  'Facet'
		TEvtString 'Name'
	end)

	---------------------------
	if mmver ~= 6 then
		MakeCmd("ShowMovie", 0x0C, function(define)
			define
			.u1  'DoubleSize'
			.b1  'unk'  -- use 0 most of time
			TEvtString 'Name'
		end)
	else
		MakeCmd("SetTextureOutdoors", 0x0C, function(define)
			define
			.u4  'Model'
			.u4  'Facet'
			TEvtString 'Name'
		end)
	end

	---------------------------
	MakeCmd("SetSprite", 0x0D, function(define)
		define
		.i4  'SpriteId'
		.u1  'Visible'  -- bit 0x20 of sprite
		TEvtString 'Name' -- in case of "0" sprite is removed (?)
	end)

	---------------------------
	MakeCmd("Cmp", 0x0E, function(define)
		EvtVar  'VarNum'  -- variable number
		.i4  'Value'  -- compare with
		YJump  '  jump'
	end)

	---------------------------
	MakeCmd("SetDoorState", 0x0F, function(define)
		define
		.u1  'Id'
		.u1  'State'  -- 00 - off, 01 - on, 02 - switch state if it isn't moving, 03 - switch state
	end)

	---------------------------
	MakeCmd("Add", 0x10, function(define)
		EvtVar  'VarNum'
		.i4  'Value'
	end)

	---------------------------
	MakeCmd("Subtract", 0x11, function(define)
		EvtVar  'VarNum'
		.i4  'Value'
	end)
	evt.Sub = evt.Subtract

	---------------------------
	MakeCmd("Set", 0x12, function(define)
		EvtVar  'VarNum'
		.i4  'Value'
	end)

	---------------------------
	MakeCmd("SummonMonsters", 0x13, function(define)
		define
		.u1  'TypeIndexInMapStats'
		.u1  'Level'
		.u1  'Count'
		.i4  'X'
		.i4  'Y'
		.i4  'Z'
		if mmver ~= 6 then
			define
			.i4  'NPCGroup'
			.i4  'unk'
		end
	end)

	---------------------------
	MakeCmd("Cmd14", 0x14, function(define)
		define
		.i4  'unk_1'
		.u1  'unk_2'
	end, true)

	---------------------------
	MakeCmd("CastSpell", 0x15, function(define)
		define
		.u1  'Number'
		.u1  'SkillLevel'  -- 0 - 3
		.u1  'Skill'
		.i4  'FromX'
		.i4  'FromY'
		.i4  'FromZ'
		.i4  'ToX'
		.i4  'ToY'
		.i4  'ToZ'
	end)

	---------------------------
	MakeCmd("SpeakNPC", 0x16, function(define)
		define
		.i4  'NPC'
	end)

	---------------------------
	MakeCmd("SetFacetBit", 0x17, function(define)
		define
		.i4  'Id'  -- id of facets group in MM7-8
		.u4  'Bit'
		.b1  'On'
	end)
	BitsArray[0x17] = 'FacetBits'

	---------------------------
	if mmver == 6 then
		MakeCmd("SetFacetBitOutdoors", 0x18, function(define)
			define
			.i4  'Model'  -- number of structure in map (those shown on left click in MapViewer)
			.i4  'Facet'  -- -1 = for all faces in structure
			.i4  'Bit'
			.b1  'On'
		end)
		BitsArray[0x18] = 'FacetBits'
	else
		MakeCmd("SetMonsterBit", 0x18, function(define)
			define
			.i4  'Monster'
			.i4  'Bit'
			.b1  'On'
		end)
		BitsArray[0x18] = 'MonsterBits'
	end

	---------------------------
	MakeCmd("RandomGoTo", 0x19, function(define)
		define
		.u1  'jumpA'
		.u1  'jumpB'
		.u1  'jumpC'
		.u1  'jumpD'
		.u1  'jumpE'
		.u1  'jumpF'
		-- 0 to skip a label
	end, true)

	---------------------------
	MakeCmd("Question", 0x1A, function(define)
		define
		.i4  'Question'
		.i4  'Answer1'
		.i4  'Answer2'
		YJump  '  jump(ok)'
	end)

	---------------------------
	MakeCmd("Cmd1B", 0x1B, function(define)
		define
		.u1  'unk_1'
		.u1  'unk_2'
	end, true)

	---------------------------
	MakeCmd("Cmd1C", 0x1C, function(define)
		define
		.u1  'unk_1'
	end, true)

	---------------------------
	MakeCmd("StatusText", 0x1D, function(define)
		define
		.i4  'Str'
	end)

	---------------------------
	MakeCmd("SetMessage", 0x1E, function(define)
		define
		.i4  'Str'
	end)

	---------------------------
	local function DeclareTimer(define)
		define
		.b1  'EachYear'
		.b1  'EachMonth'
		.b1  'EachWeek'  -- else each day
		.u1  'StartHour'
		.u1  'StartMinute'
		.u1  'StartSecond'
		.u2  'IntervalInHalfMinutes'
		.skip(2)
	end
	
	MakeCmd("OnTimer", 0x1F, DeclareTimer, true)

	---------------------------
	MakeCmd("SetLight", 0x20, function(define)
		define
		.i4  'Id'  -- light group id in MM7-8
		.b1  'On'
	end)

	---------------------------
	MakeCmd("SimpleMessage", 0x21, function(define)
		define
		.skip(1)
	end)

	---------------------------
	MakeCmd("SummonObject", 0x22, function(define)
		define
		.i4  'Type'  -- dobjlist.bin
		.i4  'X'
		.i4  'Y'
		.i4  'Z'
		.i4  'Speed'
		.u1  'Count'
		.b1  'RandomAngle'
	end)

	---------------------------
	MakeCmd("ForPlayer", 0x23, function(define)
		define
		.u1  'Player'
	end, true)

	---------------------------
	MakeCmd("GoTo", 0x24, function(define)
		YJump  'jump'
	end, true)

	---------------------------
	MakeCmd("OnLoadMap", 0x25, function(define)
		define
		.skip(1)
	end, true)

	---------------------------
	MakeCmd("OnLongTimer", 0x26, DeclareTimer, true)

	---------------------------
	MakeCmd("SetNPCTopic", 0x27, function(define)
		define
		.i4  'NPC'
		.u1  'Index'
		.i4  'Event'
	end)

	---------------------------
	MakeCmd("MoveNPC", 0x28, function(define)
		define
		.u4  'NPC'
		.u4  'HouseId' -- 2DEvents
	end)

	---------------------------
	MakeCmd("GiveItem", 0x29, function(define)
		define
		.u1  'Strength'  -- 1-6 (like described in the end of STDITEMS.TXT)
		.u1  'Type'
		.u4  'Id'
		-- If Id is 0, the item is chosen by random from the specified class with specified strength,
		-- otherwise, ItemType and ItemStremgth determine the enchancements
		-- If both unk_1 and unk_2 are 0, the item is identified and has no enchancements
	end)

	---------------------------
	MakeCmd("ChangeEvent", 0x2A, function(define)
		define
		.u4  'NewEvent'  -- (global event) for barrels etc.
	end)

	---------------------------
	MakeCmd("CheckSkill", 0x2B, function(define)
		define
		.u1  'SkillType'
		.u1  'SkillMastery'
		.i4  'OrSkillLevel'  -- Includes "Double effect" enchantments and NPC bonuses
		YJump  '  jump(>=)'
	end)

	--------------------------- MM7

	if mmver < 7 then
		return
	end

	---------------------------
	MakeCmd("CanShowTopic.Cmp", 0x2C, function(define)
		EvtVar  'VarNum'
		.i4  'Value'
		YJump  '  jump'
	end, true)

	---------------------------
	MakeCmd("CanShowTopic.Exit", 0x2D, function(define)
		define
		.skip(1)
	end, true)

	---------------------------
	MakeCmd("CanShowTopic.Set", 0x2E, function(define)
		define
		.b1  'Visible'
	end, true)

	---------------------------
	MakeCmd("SetNPCGroupNews", 0x2F, function(define)
		define
		.u4  'NPCGroup'
		.u4  'NPCNews'
	end)

	---------------------------
	MakeCmd("SetMonsterGroup", 0x30, function(define)
		define
		.u4  'Monster'
		.u4  'Group'
	end)

	---------------------------
	MakeCmd("SetNPCItem", 0x31, function(define)
		define
		.i4  'NPC'
		.i4  'Item'
		.b1  'On'
	end)

	---------------------------
	MakeCmd("SetNPCGreeting", 0x32, function(define)
		define
		.i4  'NPC'
		.i4  'Greeting'
	end)

	---------------------------
	local function CheckMonstersKilled(define)
		define
		.u1  'CheckType'  -- 0 = any monster, 1 - in group, 2 - of type, 3 - specific monster, 4 - specific monster by name (MM8)
		.u4  'Id'  -- 0 - not used, 1 - group id, 2 - monster type, 3 - monster id, 4 - id in placemon.txt (MM8 only)
		.u1  'Count'  -- 0 - all must be killed, else a number of monsters that must be killed
		if mmver == 8 then
			define.u1  'InvisibleAsDead'  -- a monster can be invisible, like pirates in Ravenshore in MM8 before you enter Regna
		end
		YJump  '  jump(>=)'
	end
	
	MakeCmd("CheckMonstersKilled", 0x33, CheckMonstersKilled)

	---------------------------
	MakeCmd("CanShowTopic.CheckMonstersKilled", 0x34, CheckMonstersKilled, true)

	---------------------------
	MakeCmd("OnLeaveMap", 0x35, function(define)
		define
		.skip(1)
	end, true)

	---------------------------
	MakeCmd("ChangeGroupToGroup", 0x36, function(define)
		define
		.i4  'Old'
		.i4  'New'
	end, true)
	
	---------------------------
	MakeCmd("ChangeGroupAlly", 0x37, function(define)
		define
		.i4  'Group'
		.i4  'Ally'
	end, true)
	
	---------------------------
	MakeCmd("CheckSeason", 0x38, function(define)
		define
		.u1  'Season'
		YJump  '  jump(ok)'
	end)

	---------------------------
	MakeCmd("SetMonGroupBit", 0x39, function(define)
		define
		.i4  'NPCGroup'
		.u4  'Bit'
		.b1  'On'
	end)
	BitsArray[0x39] = 'MonsterBits'

	---------------------------
	MakeCmd("SetChestBit", 0x3A, function(define)
		define
		.i4  'ChestId'
		.u4  'Bit'
		.b1  'On'
	end)
	BitsArray[0x3A] = 'ChestBits'

	---------------------------
	MakeCmd("FaceAnimation", 0x3B, function(define)
		define
		.u1  'Player'
		.u1  'Animation'
	end)

	---------------------------
	MakeCmd("SetMonsterItem", 0x3C, function(define)
		define
		.i4  'Monster'
		.i4  'Item'
		.b1  'Has'
	end)

	--------------------------- MM8

	if mmver < 8 then
		return
	end

	---------------------------
	MakeCmd("OnDateTimer", 0x3D, function(define)
		define
		.u1  'Id'  -- 0x3D00 is added to this
		.u1  'HourPlus1'
		.u1  'DayPlus1'
		.u1  'WeekPlus1'
		.u1  'YearPlus1'
		.u2  'FullYearPlus1'
		.u1  'EnableAfter'  -- works every half of second
		-- By default date timer is disabled!
	end, true)

	---------------------------
	MakeCmd("EnableDateTimer", 0x3E, function(define)
		define
		.u2  'IdPlus0x3D00'  -- add 0x3D00 to Id of Cmd_3D
		.b1  'On'
	end, true)

	---------------------------
	MakeCmd("StopDoor", 0x3F, function(define)
		define
		.i4  'Id'
	end)

	---------------------------
	MakeCmd("CheckItemsCount", 0x40, function(define)
		define
		.u2  'MinItemIndex'
		.u2  'MaxItemIndex'
		.u2  'Count'
		YJump  '  jump(>=)'
	end)

	---------------------------
	MakeCmd("RemoveItems", 0x41, function(define)
		define
		.u2  'MinItemIndex'
		.u2  'MaxItemIndex'
		.u2  'Count'
	end)

	---------------------------
	MakeCmd("Jump", 0x42, function(define)
		define
		.i2  'Direction'
		.i2  'ZAngle'
		.i2  'Speed'
		.i2  'unk'
	end)

	---------------------------
	MakeCmd("IsTotalBountyInRange", 0x43, function(define)
		define
		.i4  'MinGold'
		.i4  'MaxGold'
		YJump  '  jump(ok)'
	end)

	---------------------------
	MakeCmd("IsPlayerInParty", 0x44, function(define)
		define
		.u4  'Id' -- Roster.txt
		YJump  '  jump(ok)'
	end)

end)()

----------- Enums

local SkillNamesList = {}
for k, v in pairs(const.Skills) do
	SkillNamesList[v + 1] = k
end

if mmver ~= 8 then
	evt.Players = {
		[0] = 0,
		[1] = 1,
		[2] = 2,
		[3] = 3,
		Current = 4,
		current = 4,
		All = 5,
		all = 5,
		Random = 6,
		random = 6,
	}
else
	evt.Players = {
		[0] = 0,
		[1] = 1,
		[2] = 2,
		[3] = 3,
		[4] = 4,
		All = 5,
		Random = 6,
		Current = 7,
	}
end

PlayerToStr = {}

for k, v in pairs(evt.Players) do
	if type(k) ~= "number" then
		PlayerToStr[v] = "Players."..k
	end
end


do
	local p = 1
	evt.VarNum = {}
	local function add(name)
		if name then
			evt.VarNum[name] = p
		end
		p = p + 1
	end
	
	add("SexIs")                     -- 01
	add("ClassIs")                   -- 02
	add("HP")                        -- 03
	add("HasFullHP")                 -- 04
	add("SP")                        -- 05
	add("HasFullSP")                 -- 06
	add("ArmorClass")                -- 07
	add("ArmorClassBonus")           -- 08
	add("BaseLevel")                 -- 09
	add("LevelBonus")                -- 0A
	add("AgeBonus")                  -- 0B
	add("Awards")                    -- 0C
	add("Experience")                -- 0D
	add(nil)                         -- 0E : not used
	add(nil)                         -- 0F : not used
	add("QBits")                     -- 10
	add("Inventory")                 -- 11
	add("HourIs")                    -- 12
	add("DayOfYearIs")               -- 13
	add("DayOfWeekIs")               -- 14
	add("Gold")                      -- 15
	add("GoldAddRandom")             -- 16
	add("Food")                      -- 17
	add("FoodAddRandom")             -- 18
	add("MightBonus")                -- 19
	add("IntellectBonus")            -- 1A
	add("PersonalityBonus")          -- 1B
	add("EnduranceBonus")            -- 1C
	add("SpeedBonus")                -- 1D
	add("AccuracyBonus")             -- 1E
	add("LuckBonus")                 -- 1F
	add("BaseMight")                 -- 20
	add("BaseIntellect")             -- 21
	add("BasePersonality")           -- 22
	add("BaseEndurance")             -- 23
	add("BaseSpeed")                 -- 24
	add("BaseAccuracy")              -- 25
	add("BaseLuck")                  -- 26
	add("CurrentMight")              -- 27
	add("CurrentIntellect")          -- 28
	add("CurrentPersonality")        -- 29
	add("CurrentEndurance")          -- 2A
	add("CurrentSpeed")              -- 2B
	add("CurrentAccuracy")           -- 2C
	add("CurrentLuck")               -- 2D
	if mmver == 6 then
		add("FireResistance")          -- 2E
		add("ElecResistance")          -- 2F
		add("ColdResistance")          -- 30
		add("PoisonResistance")        -- 31
		add("MagicResistance")         -- 32
		add("FireResBonus")            -- 33
		add("ElecResBonus")            -- 34
		add("ColdResBonus")            -- 35
		add("PoisonResBonus")          -- 36
		add("MagicResBonus")           -- 37
	else
		add("FireResistance")          -- 2E
		add("AirResistance")           -- 2F
		add("WaterResistance")         -- 30
		add("EarthResistance")         -- 31
		add("SpiritResistance")        -- 32
		add("MindResistance")          -- 33
		add("BodyResistance")          -- 34
		add(nil)                       -- 35 : LightResistance
		add(nil)                       -- 36 : DarkResistance
		add(nil)                       -- 37 : PhysResistance
		add(nil)                       -- 38 : MagicResistance
		add("FireResBonus")            -- 39
		add("AirResBonus")             -- 3A
		add("WaterResBonus")           -- 3B
		add("EarthResBonus")           -- 3C
		add("SpiritResBonus")          -- 3D
		add("MindResBonus")            -- 3E
		add("BodyResBonus")            -- 3F
		add(nil)                       -- 40 : LightResBonus
		add(nil)                       -- 41 : DarkResBonus
		add(nil)                       -- 42 : PhysResistance
		add(nil)                       -- 43 : MagicResistance
	end
	evt.VarNum.Skills = {}           -- 38-56/44-68/44-6A
	for i, v in ipairs(SkillNamesList) do
		evt.VarNum.Skills[i - 1] = p
		add(v.."Skill")
	end
	add("Cursed")                    -- 57/69/6B
	add("Weak")                      -- 58/6A
	add("Asleep")                    -- 59/6B
	add("Afraid")                    -- 5A/6C
	add("Drunk")                     -- 5B/6D
	add("Insane")                    -- 5C/6E
	add("PoisonedGreen")             -- 5D/6F
	add("DiseasedGreen")             -- 5E/70
	add("PoisonedYellow")            -- 5F/71
	add("DiseasedYellow")            -- 60/72
	add("PoisonedRed")               -- 61/73
	add("DiseasedRed")               -- 62/74
	add("Paralysed")                 -- 63/75
	add("Unconsious")                -- 64/76
	add("Dead")                      -- 65/77
	add("Stoned")                    -- 66/78
	add("Eradicated")                -- 67/79
	add("MainCondition")             -- 68/7A/7C
	evt.VarNum.MapVars = {}          -- 69-CC/7B-DE/7D-E0
	for i = 0, 99 do
		evt.VarNum.MapVars[i] = p
		add("MapVar"..i)
	end
	add("AutonotesBits")             -- CD/DF/E1
	add("IsMightMoreThanBase")       -- CE/E0
	add("IsIntellectMoreThanBase")   -- CF/E1
	add("IsPersonalityMoreThanBase") -- D0/E2
	add("IsEnduranceMoreThanBase")   -- D1/E3
	add("IsSpeedMoreThanBase")       -- D2/E4
	add("IsAccuracyMoreThanBase")    -- D3/E5
	add("IsLuckMoreThanBase")        -- D4/E6
	add("PlayerBits")                -- D5/E7/E9
	add(mmver ~= 8 and "NPCs" or nil)-- D6/E8
	add("ReputationIs")              -- D7/E9
	for i = 0, 5 do  add(nil)  end   -- D8-DD/... : something time-related in 0x90E85C/...
	add("Flying")                    -- DE/F0/F2
	add("HasNPCProfession")          -- DF/F1
	add("TotalCircusPrize")          -- E0/F2
	add("SkillPoints")               -- E1/F3
	add("MonthIs")                   -- E2/F4
	if mmver >= 7 then
		add("Counter1")                -- F5/F7
		add("Counter2")                -- F6
		add("Counter3")                -- F7
		add("Counter4")                -- F8
		add("Counter5")                -- F9
		add("Counter6")                -- FA
		add("Counter7")                -- FB
		add("Counter8")                -- FC
		add("Counter9")                -- FD
		add("Counter10")               -- FE/100
		for i = 0x101, 0x114 do
			add(nil)                     -- FF-112/101-114 : set something to current time and play sound (not used by CMP)
			                             -- (no metter what value you use) (the value set is never used)
		end
		add("Reputation")              -- 113/115
		evt.VarNum.History = {}        -- 114/116
		for i = 1, 29 do
			evt.VarNum.History[i] = p
			add("History"..i)
		end
		add("MapAlert")                -- 131/133
		add("BankGold")                -- 132/134
		add("Deaths")                  -- 133/135
		add("MontersHunted")           -- 134/136
		add("PritsonTerms")            -- 135/137
		add("ArenaWinsPage")           -- 136/138
		add("ArenaWinsSquire")         -- 137/138
		add("ArenaWinsKnight")         -- 138/13A
		add("ArenaWinsLord")           -- 139/13B
		add("Invisible")               -- 13A/13C
		add("IsWearingItem")           -- 13B/13D
		if mmver == 8 then
			add("Players")               -- 13E
		end
	end
	
	if mmver == 6 and evt.VarNum.MonthIs ~= 0xE2 or mmver == 7 and evt.VarNum.IsWearingItem ~= 0x13B or mmver == 8 and evt.VarNum.Players ~= 0x13E then
		error("wrong evt.VarNum")
	end
end

VarNumToStr = {}

for k, v in pairs(evt.VarNum) do
	VarNumToStr[v] = '"'..k..'"' --"VarNum."..k
end
